package org.xbib.interlibrary.elasticsearch.sru;

import com.fasterxml.jackson.dataformat.xml.XmlFactory;
import com.fasterxml.jackson.dataformat.xml.ser.ToXmlGenerator;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Templates;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.aggregations.Aggregation;
import org.elasticsearch.search.aggregations.AggregationBuilder;
import org.elasticsearch.search.aggregations.bucket.terms.LongTerms;
import org.elasticsearch.search.aggregations.bucket.terms.StringTerms;
import org.elasticsearch.search.aggregations.bucket.terms.Terms;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.sort.SortBuilder;
import org.xbib.content.XContentBuilder;
import org.xbib.content.xml.XmlNamespaceContext;
import org.xbib.content.xml.XmlXContent;
import org.xbib.content.xml.XmlXParams;
import org.xbib.cql.CQLParser;
import org.xbib.cql.elasticsearch.ElasticsearchQueryGenerator;
import org.xbib.datastructures.api.Builder;
import org.xbib.datastructures.json.tiny.JsonBuilder;
import org.xbib.elx.api.SearchClient;
import org.xbib.elx.api.SearchDocument;
import org.xbib.elx.api.SearchResult;
import org.xbib.interlibrary.action.sru.AbstractSearchRetrieveService;
import org.xbib.interlibrary.action.sru.Bucket;
import org.xbib.interlibrary.action.sru.ClasspathUriResolver;
import org.xbib.interlibrary.action.sru.Constants;
import org.xbib.interlibrary.action.sru.Facet;
import org.xbib.interlibrary.action.sru.Record;
import org.xbib.interlibrary.action.sru.SearchRetrieveRequest;
import org.xbib.interlibrary.action.sru.SearchRetrieveResponse;
import org.xbib.interlibrary.action.sru.SearchRetrieveResponseBuilder;
import org.xbib.interlibrary.action.sru.SearchRetrieveService;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.api.action.ServiceArguments;

public class ElasticsearchSearchRetrieveService extends AbstractSearchRetrieveService<SearchRetrieveRequest, SearchRetrieveResponse, SearchRetrieveService> {

    private static final Logger logger = Logger.getLogger(ElasticsearchSearchRetrieveService.class.getName());

    private static final String KEY = "identifier";

    private final SearchClient searchClient;

    private final AtomicBoolean enabled;

    private final AtomicLong counter;

    private final String globalField;

    public ElasticsearchSearchRetrieveService(ServiceArguments arguments, SearchClient searchClient) {
        super(arguments);
        this.enabled = new AtomicBoolean(arguments.getServiceSettingsAsBoolean("enabled", true));
        this.counter = new AtomicLong();
        this.searchClient = searchClient;
        this.globalField = arguments.getServiceSettings("globalField", "cql.any");
    }

    @Override
    public String getName() {
        return arguments.getName();
    }

    @Override
    public String getLabel() {
        return arguments.getName().toUpperCase(Locale.ROOT);
    }

    @Override
    public void execute(SearchRetrieveRequest searchRetrieveRequest, SearchRetrieveResponseBuilder responseBuilder) throws IOException {
        if (!enabled.get()) {
            return;
        }
        String cqlQuery = searchRetrieveRequest.getSearchString();
        boolean matchAll = cqlQuery == null || cqlQuery.trim().isEmpty() || cqlQuery.equals(".");
        ElasticsearchFacetsGenerator facetsGenerator = new ElasticsearchFacetsGenerator();
        ElasticsearchSortGenerator sortGenerator = new ElasticsearchSortGenerator();
        ElasticsearchQueryGenerator generator =
                new ElasticsearchQueryGenerator(globalField, null, null, facetsGenerator, sortGenerator, false);
        createElasticsearchQueryFromCQL(generator, searchRetrieveRequest);
        String endpoint = searchRetrieveRequest.getEndPoint() != null ?
                searchRetrieveRequest.getEndPoint() : arguments.getServiceSettings("index", null);
        Builder logmsg = JsonBuilder.builder();
        long l = counter.incrementAndGet();
        logmsg.beginMap()
                .field("id", l)
                .field("type", "request")
                .fieldIfNotNull("endpoint", searchRetrieveRequest.getEndPoint())
                .field("version", searchRetrieveRequest.getVersion())
                .field("operation", Constants.SEARCH_RETRIEVE_OPERATION)
                .fieldIfNotNull("recordSchema", searchRetrieveRequest.getRecordSchema())
                .fieldIfNotNull("extraRequestData", searchRetrieveRequest.getExtraRequestData())
                .field("query", searchRetrieveRequest.getSearchString())
                .fieldIfNotNull("facetLimit", searchRetrieveRequest.getFacetLimit())
                .endMap();
        logger.log(Level.INFO, logmsg.build());
        Optional<SearchResult> optionalSearchResult;
        if (matchAll) {
            optionalSearchResult = searchClient.search(searchRequestBuilder -> searchRequestBuilder
                    .setIndices(endpoint.split(","))
                    .setQuery(QueryBuilders.matchAllQuery()));
        } else {
            SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();
            searchSourceBuilder.from(searchRetrieveRequest.getOffset());
            searchSourceBuilder.size(searchRetrieveRequest.getSize());
            searchSourceBuilder.query(QueryBuilders.wrapperQuery(generator.getQueryResult()));
            for (SortBuilder<?> builder : sortGenerator.getList()) {
                searchSourceBuilder.sort(builder);
            }
            for (AggregationBuilder builder : facetsGenerator.getList()) {
                searchSourceBuilder.aggregation(builder);
            }
            logger.log(Level.FINE, "search source = " + searchSourceBuilder);
            optionalSearchResult = searchClient.search(searchRequestBuilder -> searchRequestBuilder
                    .setIndices(endpoint.split(","))
                    .setSource(searchSourceBuilder));
        }
        if (optionalSearchResult.isPresent()) {
            SearchResult searchResult = optionalSearchResult.get();
            logmsg = JsonBuilder.builder();
            logmsg.beginMap()
                    .field("id", l)
                    .field("type", "response")
                    .fieldIfNotNull("endpoint", endpoint)
                    .field("version", searchRetrieveRequest.getVersion())
                    .field("operation", Constants.SEARCH_RETRIEVE_OPERATION)
                    .field("total", searchResult.getTotal())
                    .field("took", searchResult.getTook())
                    .endMap();
            logger.log(Level.INFO, logmsg.build());
            buildSearchRetrieveResponse(searchRetrieveRequest, responseBuilder, endpoint, searchResult, endpoint);
        } else {
            logmsg = JsonBuilder.builder();
            logmsg.beginMap()
                    .field("id", l)
                    .field("type", "response")
                    .fieldIfNotNull("endpoint", endpoint)
                    .field("version", searchRetrieveRequest.getVersion())
                    .field("operation", Constants.SEARCH_RETRIEVE_OPERATION)
                    .field("empty", true)
                    .endMap();
            logger.log(Level.INFO, logmsg.build());
            responseBuilder.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
        }
    }

    private void createElasticsearchQueryFromCQL(ElasticsearchQueryGenerator generator, SearchRetrieveRequest searchRequest) {
        String cqlQuery = searchRequest.getSearchString();
        generator.setFrom(searchRequest.getOffset());
        generator.setSize(searchRequest.getSize());
        if (searchRequest.getFilter() != null) {
            generator.filter(searchRequest.getFilter());
        }
        Map<String, Collection<String>> orFilter = searchRequest.getOrFilter();
        if (orFilter != null) {
            for (Map.Entry<String, Collection<String>> me : orFilter.entrySet()) {
                String key = me.getKey();
                if (key.startsWith("filter.or.")) {
                    key = key.substring(10);
                }
                generator.orfilter(key, me.getValue());
            }
        }
        Map<String, Collection<String>> andFilter = searchRequest.getAndFilter();
        if (andFilter != null) {
            for (Map.Entry<String, Collection<String>> me : andFilter.entrySet()) {
                String key = me.getKey();
                if (key.startsWith("filter.and.")) {
                    key = key.substring(11);
                }
                generator.andfilter(key, me.getValue());
            }
        }
        if (searchRequest.getFacetLimit() != null) {
            generator.facet(searchRequest.getFacetLimit(), searchRequest.getFacetSort());
        }
        CQLParser parser = new CQLParser(cqlQuery);
        parser.parse();
        parser.getCQLQuery().accept(generator);
    }

    private void buildSearchRetrieveResponse(SearchRetrieveRequest request,
                                             SearchRetrieveResponseBuilder response,
                                             String index,
                                             SearchResult searchResult,
                                             String... values) throws IOException {
        List<Record> records = new LinkedList<>();
        List<Facet> facets = new LinkedList<>();
        if (!searchResult.getDocuments().isEmpty()) {
            String relatedIndex = request.getExtraRequestData() != null ? request.getExtraRequestData() : "holdings";
            List<Builder> logmsgs = new LinkedList<>();
            for (SearchDocument doc : searchResult.getDocuments()) {
                String recordIdentifier = doc.getIndex().replaceAll("\\d+$", "") + "/" + escape(doc.getId());
                String recordSchema = request.getRecordSchema() != null ? request.getRecordSchema() : "mods";
                String recordData;
                XmlXParams xmlParams = createXmlXContentBuilderForSRU(request.getVersion(), "source");
                XContentBuilder builder = XmlXContent.contentBuilder(xmlParams);
                builder.startObject();
                filter(builder, doc.getFields(), KEY, values);
                fetchRelatedDocs(request, builder, logmsgs, index, doc.getIndex(), relatedIndex, doc.getId(), values);
                builder.endObject();
                recordData = request.getStyleSheet() == null && request.getDefaultStyleSheet() == null ? builder.string() :
                        transform(ElasticsearchSearchRetrieveService.class.getClassLoader(), request.getStyleSheet(), request.getDefaultStyleSheet(), builder.string());
                records.add(new Record(recordIdentifier, recordSchema, "xml", recordData));
            }
            if (!logmsgs.isEmpty()) {
                Builder logBuilder = JsonBuilder.builder();
                logBuilder.beginMap();
                logBuilder.field("type", "related");
                logBuilder.field("index", index);
                logBuilder.fieldIfNotNull("recordschema", request.getRecordSchema());
                logBuilder.beginCollection("related");
                for (Builder b : logmsgs) {
                    logBuilder.buildValue(b.build());
                }
                logBuilder.endCollection();
                logger.log(Level.INFO, logBuilder.build());
            }
        }
        if (searchResult.getAggregations() != null) {
            String query = request.getSearchString();
            for (Aggregation aggregation : searchResult.getAggregations()) {
                String facetname;
                if (aggregation instanceof StringTerms) {
                    StringTerms stringTerms = (StringTerms) aggregation;
                    facetname =  stringTerms.getName();
                    List<Bucket> buckets = new LinkedList<>();
                    for (Terms.Bucket bucket : stringTerms.getBuckets()) {
                        String bucketTerm = bucket.getKey().toString();
                        long bucketCount = bucket.getDocCount();
                        String newFilter = stringTerms.getName() + "=\"" + URLEncoder.encode(bucket.getKey().toString(), StandardCharsets.UTF_8) + "\"";
                        String filter = request.getFilter() != null ? request.getFilter() + " and " + newFilter : newFilter;
                        String path = (request.getPath() != null ? request.getPath() : "") +
                                "?operation=searchRetrieve" +
                                (request.getVersion() != null ? "&version=" + request.getVersion() : "") +
                                (query != null ? "&query=" + query : "") +
                                ("&filter=" + filter) +
                                (request.getRecordSchema() != null ? "&recordSchema=" + request.getRecordSchema() : "") +
                                (request.getExtraRequestData() != null ? "&extraRequestData=" + request.getExtraRequestData() : "") +
                                (request.getOffset() != null ? "&startRecord=" + request.getOffset() : "") +
                                (request.getSize() != null ? "&maximumRecords=" + request.getSize() : "") +
                                (request.getFacetLimit() != null ? "&facetLimit=" + request.getFacetLimit() : "");
                        String bucketRequestUrl = escape(path).toString();
                        buckets.add(new Bucket(bucketTerm, bucketRequestUrl, Long.toString(bucketCount)));
                    }
                    facets.add(new Facet(facetname, buckets));
                } else if (aggregation instanceof LongTerms) {
                    LongTerms longTerms = (LongTerms) aggregation;
                    facetname = longTerms.getName();
                    List<Bucket> buckets = new LinkedList<>();
                    for (Terms.Bucket bucket : longTerms.getBuckets()) {
                        String bucketTerm = bucket.getKey().toString();
                        long bucketCount = bucket.getDocCount();
                        String newFilter = longTerms.getName() + "=\"" + URLEncoder.encode(bucket.getKey().toString(), StandardCharsets.UTF_8) + "\"";
                        String filter = request.getFilter() != null ? request.getFilter() + " and " + newFilter : newFilter;
                        String path = request.getPath() +
                                "?operation=searchRetrieve" +
                                (request.getVersion() != null ? "&version=" + request.getVersion() : "") +
                                (query != null ? "&query=" + query : "") +
                                ("&filter=" + filter) +
                                (request.getRecordSchema() != null ? "&recordSchema=" + request.getRecordSchema() : "") +
                                (request.getExtraRequestData() != null ? "&extraRequestData=" + request.getExtraRequestData() : "") +
                                (request.getOffset() != null ? "&startRecord=" + request.getOffset() : "") +
                                (request.getSize() != null ? "&maximumRecords=" + request.getSize() : "") +
                                (request.getFacetLimit() != null ? "&facetLimit=" + request.getFacetLimit() : "");
                        String bucketRequestUrl = escape(path).toString();
                        buckets.add(new Bucket(bucketTerm, bucketRequestUrl, Long.toString(bucketCount)));
                    }
                    facets.add(new Facet(facetname, buckets));
                } else {
                    logger.log(Level.WARNING, "unknown aggregation type: " + aggregation.getClass());
                }
            }
        }
        response.setRequest(request);
        response.setTotal(searchResult.getTotal());
        response.setRecords(records);
        response.setFacets(facets);
        response.setStatus(getName(), Response.Status.OK);
    }

    private void fetchRelatedDocs(SearchRetrieveRequest request,
                                  XContentBuilder builder,
                                  List<Builder> logmsgs,
                                  String index,
                                  String hitIndex,
                                  String relatedIndex,
                                  String uid,
                                  String... values) throws IOException {
        if (relatedIndex == null) {
            return;
        }
        String relatedHitIndex = hitIndex.replaceAll("\\d{8,}$","") + relatedIndex;
        Stream<SearchDocument> stream = searchClient.search(searchRequestBuilder ->
                searchRequestBuilder.setIndices(relatedHitIndex)
                        .setQuery(QueryBuilders.termQuery("xbib.uid", uid)), TimeValue.timeValueMinutes(1), 100);
        builder.startArray(relatedIndex);
        stream.forEach(doc -> {
            try {
                XmlXParams holdingsXmlXParams = createXmlXContentBuilderForSRU(request.getVersion(), "holdings");
                XContentBuilder hitBuilder = XmlXContent.contentBuilder(holdingsXmlXParams);
                hitBuilder.startObject();
                hitBuilder.field("index", escape(doc.getIndex()));
                hitBuilder.field("id", escape(doc.getId()));
                hitBuilder.startObject("source");
                boolean isLocal = isLocal(doc.getIndex(), index);
                boolean hasContent = filter(hitBuilder, doc.getFields(), isLocal ? null : KEY, values);
                hitBuilder.endObject();
                hitBuilder.endObject();
                if (hasContent) {
                    builder.copy(hitBuilder);
                }
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
        builder.endArray();
        Builder logBuilder = JsonBuilder.builder();
        logBuilder.beginMap()
                .field("index", relatedHitIndex)
                .field("uid", uid)
                .endMap();
        logmsgs.add(logBuilder);
    }

    /**
     * Filter out unwanted elements from Elasticsearch document.
     * Remove all "xbib" keys, and remove ISILs if they are set in "identifier" and are not the index
     *
     * @param builder the builder
     * @param map Elasticsearch document
     */
    @SuppressWarnings("unchecked")
    private static boolean filter(XContentBuilder builder,
                                  Map<String,Object> map,
                                  String key,
                                  String... values) throws IOException {
        boolean hasContent = false;
        for (Map.Entry<String,Object> me : map.entrySet()) {
            if ("xbib".equals(me.getKey())) {
                continue;
            }
            if (key == null) {
                builder.field(me.getKey(), me.getValue());
                hasContent = true;
                continue;
            }
            if (me.getValue() instanceof Map) {
                Map<String, Object> m = findKey((Map<String, Object>) me.getValue(), key, values);
                if (m != null) {
                    builder.field(me.getKey());
                    builder.map(m);
                    hasContent = true;
                }
            } else if (me.getValue() instanceof List) {
                builder.startArray(me.getKey());
                List<Object> list = (List<Object>) me.getValue();
                for (Object o : list) {
                    if (o instanceof Map) {
                        Map<String, Object> m = findKey((Map<String, Object>) o, key, values);
                        if (m != null) {
                            builder.map(m);
                            hasContent = true;
                        }
                    } else {
                        builder.value(o);
                        hasContent = true;
                    }
                }
                builder.endArray();
            } else {
                builder.field(me.getKey(), me.getValue());
                hasContent = true;
            }
        }
        return hasContent;
    }

    private static Map<String, Object> findKey(Map<String,Object> m, String key, String... values) {
        boolean b = false;
        if (m.containsKey(key)) {
            for (String v : values) {
                String s = m.get(key).toString();
                if (s.indexOf("-") < s.lastIndexOf("-")) {
                    if (s.startsWith(v)) {
                        b = true;
                    }
                } else if (v.equals(s)) {
                    b = true;
                }
            }
        } else {
            b = true;
        }
        return b ? m : null;
    }

    private static boolean isLocal(String index, String otherIndex) {
        return index.replaceAll("\\d{8,}$","").startsWith(otherIndex.replaceAll("\\d{8,}$",""));
    }

    private static CharSequence escape(CharSequence xml) {
        return org.xbib.content.xml.util.XMLUtil.escape(xml);
    }

    @Override
    public void close() {
    }

    @Override
    public int compareTo(Service<SearchRetrieveRequest, SearchRetrieveResponse, SearchRetrieveResponseBuilder> o) {
        return getName().compareTo(o.getName());
    }

    public static XmlXParams createXmlXContentBuilderForSRU(String version, String elementName) {
        XmlNamespaceContext context = XmlNamespaceContext.newInstance();
        context.addNamespace("", NS_URI);
        if ("1.1".equals(version) || "1.2".equals(version)) {
            context.addNamespace("sru", Constants.SRW_NS_URI);
        } else {
            context.addNamespace("sru", Constants.SEARCH_RETRIEVE_NS_URI);
        }
        QName qName = new QName(NS_URI, elementName);
        return new XmlXParams(qName, context, createXmlFactory());
    }

    private static XmlFactory createXmlFactory() {
        XMLInputFactory inputFactory = createXMLInputFactory();
        XMLOutputFactory outputFactory = createXMLOutputFactory();
        XmlFactory xmlFactory = new XmlFactory(inputFactory, outputFactory);
        xmlFactory.configure(ToXmlGenerator.Feature.WRITE_XML_1_1, true);
        return xmlFactory;
    }

    private static XMLInputFactory createXMLInputFactory() {
        XMLInputFactory inputFactory = XMLInputFactory.newFactory();
        inputFactory.setProperty("javax.xml.stream.isNamespaceAware", Boolean.TRUE);
        inputFactory.setProperty("javax.xml.stream.isValidating", Boolean.FALSE);
        inputFactory.setProperty("javax.xml.stream.isCoalescing", Boolean.TRUE);
        inputFactory.setProperty("javax.xml.stream.isReplacingEntityReferences", Boolean.FALSE);
        inputFactory.setProperty("javax.xml.stream.isSupportingExternalEntities", Boolean.FALSE);
        return inputFactory;
    }

    private static XMLOutputFactory createXMLOutputFactory() {
        XMLOutputFactory outputFactory = XMLOutputFactory.newFactory();
        outputFactory.setProperty("javax.xml.stream.isRepairingNamespaces", Boolean.FALSE);
        outputFactory.setProperty("com.ctc.wstx.useDoubleQuotesInXmlDecl", Boolean.TRUE);
        return outputFactory;
    }

    private static SAXTransformerFactory saxTransformerFactory;

    private static String transform(ClassLoader cl,
                                   String stylesheet,
                                   String defaultStylesheet,
                                   String renderedTemplate)
            throws IOException {
        if (saxTransformerFactory == null) {
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            transformerFactory.setURIResolver(new ClasspathUriResolver(cl));
            saxTransformerFactory = (SAXTransformerFactory) transformerFactory;
            Logger.getLogger(ElasticsearchSearchRetrieveService.class.getName()).log(Level.INFO, "new transformer factory = " + transformerFactory.getClass().getName());
        }
        StreamSource xslSource;
        try {
            xslSource = (StreamSource) saxTransformerFactory.getURIResolver().resolve(stylesheet, null);
        } catch (TransformerException e1) {
            try {
                xslSource = (StreamSource) saxTransformerFactory.getURIResolver().resolve(defaultStylesheet, null);
            } catch (TransformerException e2) {
                logger.log(Level.WARNING, e2.getMessage(), e2);
                return renderedTemplate;
            }
        }
        StringWriter sw = new StringWriter();
        try {
            Templates templates = saxTransformerFactory.newTemplates(xslSource);
            if (xslSource != null && xslSource.getReader() != null) {
                xslSource.getReader().close();
            }
            Result result = new StreamResult(sw);
            TransformerHandler handler = saxTransformerFactory.newTransformerHandler(templates);
            handler.setResult(result);
            saxTransformerFactory.newTransformer()
                    .transform(new StreamSource(new StringReader(renderedTemplate)), new SAXResult(handler));
        } catch (TransformerException e) {
            throw new IOException(e);
        }
        return sw.toString();
    }

    private static final String NS_URI = "http://xbib.org/ns/sru/elasticsearch/source/1.0/";

}
